# Autogenerated by Astropy-affiliated package astrowidgets's setup.py on 2019-10-14 20:41:29 UTC
from __future__ import unicode_literals
import datetime

version = "0.1.1"
githash = "9891e53701142d643276fd3f7cb1b6622fc10bd4"


major = 0
minor = 1
bugfix = 1

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 10, 14, 20, 41, 29)
debug = False

astropy_helpers_version = "3.1.dev1156"
